#include "BasisFunction.h"
#include "DOFAdmin.h"
#include "Element.h"
#include "ElInfo.h"
#include "FiniteElemSpace.h"
#include "Global.h"
#include "Mesh.h"
#include "Parameters.h"
#include "Traverse.h"

#include "CompositeFEMMethods.h"


void CompositeFEMMethods::setPosLsToVal(DOFVector<double> *dof,
					const double &val,
					const DOFVector<double> *lsFct_dof)
{
  DOFVector<double>::Iterator it_dof(dof, USED_DOFS);
  DOFVector<double>::Iterator it_lsFct_dof(
                     const_cast<DOFVector<double> *>(lsFct_dof), 
		     USED_DOFS);
  for (it_dof.reset(), it_lsFct_dof.reset();
       !it_dof.end();
       ++it_dof, ++it_lsFct_dof) {

    // Is vertex in domain with positive level set function values ?
    if (*it_lsFct_dof > 0)
      *it_dof = val;
  }
}

void CompositeFEMMethods::setPosLsToFct(DOFVector<double> *dof,
					const AbstractFunction<double, WorldVector<double> > *fct,
					const DOFVector<double> *lsFct_dof)
{
  const BasisFunction *basisFcts = dof->getFESpace()->getBasisFcts();
  const DOFAdmin *admin = dof->getFESpace()->getAdmin();
  const int dim = dof->getFESpace()->getMesh()->getDim();
  const double *locVec;
  const Element *el;
  const DegreeOfFreedom *locInd;
  TraverseStack stack;

  ElInfo *elInfo = stack.traverseFirst(dof->getFESpace()->getMesh(), -1, 
				       Mesh::CALL_LEAF_EL | 
				       Mesh::FILL_COORDS);
  
  while (elInfo) {

    el = elInfo->getElement();
    
    // Get level set function values for all vertices of element.
    locVec = lsFct_dof->getLocalVector(el, NULL);

    // Get dof indices of vertices.
    locInd = basisFcts->getLocalIndices(el, admin, NULL);

    for (int i=0; i<=dim; ++i) {
      // Is vertex in domain with positive level set function values ?
      if (locVec[i] > 0)
	(*dof)[locInd[i]] = (*fct)(elInfo->getCoord(i));
    }

    elInfo = stack.traverseNext(elInfo);
  }
}

void CompositeFEMMethods::printBoundaryElements(const std::string fn_str,
						ElementLevelSet *elLS,
						FiniteElemSpace *feSpace)
{
  FUNCNAME("CompositeFEMMethods::printBoundaryElements()");

  int dim = feSpace->getMesh()->getDim();
  std::string fn_main;
  std::string fn;
  GET_PARAMETER(0, fn_str + "->filename", &fn_main);
  fn = fn_main + "boundary_elements";
  int elStatus;

  ofstream boundaryOut(fn.c_str());

  // ===== Traverse mesh and print boundary elements. =====
  TraverseStack stack;
  int boundEl_cntr = 0;
  WorldVector<double> coord;

  const int nBasFcts = feSpace->getBasisFcts()->getNumber();
  DegreeOfFreedom *locInd = new DegreeOfFreedom[nBasFcts];

  ElInfo *loc_elInfo = stack.traverseFirst(feSpace->getMesh(),
					   -1, 
					   Mesh::CALL_LEAF_EL | 
					   Mesh::FILL_BOUND |
					   Mesh::FILL_COORDS);
  while (loc_elInfo) {

    // Get local indices of vertices.
    feSpace->getBasisFcts()->getLocalIndices(
	  const_cast<Element *>(loc_elInfo->getElement()),
	  const_cast<DOFAdmin *>(feSpace->getAdmin()),
	  locInd); 

    // Get element status.
    elStatus = elLS->createElementLevelSet(loc_elInfo);

    // Is element cut by the interface ?
    if (elStatus == ElementLevelSet::LEVEL_SET_BOUNDARY) {
      ++boundEl_cntr;

      boundaryOut << loc_elInfo->getElement()->getIndex() << ": \t";

      for (int i=0; i<=dim; ++i) {

	coord = loc_elInfo->getCoord(i);

	if (i > 0) {
	  boundaryOut << "\n\t";
	}

	boundaryOut << "\t" << coord[0];
	for (int j=1; j<dim; ++j) {
	  boundaryOut << "; " << coord[j] ;
	}
	boundaryOut << " (" << locInd[i] << ")";
      }
      boundaryOut << "\n";
    }

    loc_elInfo = stack.traverseNext(loc_elInfo);

  }  // end of: mesh traverse

  delete [] locInd;

  boundaryOut << "\nNumber of boundary elements: \t" << boundEl_cntr << "\n";

  boundaryOut.close();
}
