// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file BoundaryCondition.h */

#ifndef AMDIS_BOUNDARYCONDITION_H
#define AMDIS_BOUNDARYCONDITION_H

#include "Boundary.h"
#include "FiniteElemSpace.h"
#include "AMDiS_fwd.h"

namespace AMDiS {

  /**
   * \ingroup Assembler
   *
   * \brief
   * Sub class of BoundaryCondition. Local boundary conditions are filled
   * while mesh traversal.
   */
  class BoundaryCondition
  {
  public:
    /// Constructor.
    BoundaryCondition(BoundaryType type, 
		      const FiniteElemSpace *rowFESpace_,
		      const FiniteElemSpace *colFESpace_ = NULL) 
      : boundaryType(type),
	rowFESpace(rowFESpace_),
	colFESpace(colFESpace_)
    {
      if (!colFESpace) 
	colFESpace = rowFESpace;
    }

    /// Returns \ref boundaryType.
    inline BoundaryType getBoundaryType() 
    { 
      return boundaryType; 
    }

    /// Returns \ref rowFESpace.
    inline const FiniteElemSpace *getRowFESpace() 
    { 
      return rowFESpace; 
    }

    /// Returns \ref rowFESpace.
    inline const FiniteElemSpace *getColFESpace() 
    { 
      return colFESpace; 
    }

    virtual void initMatrix(DOFMatrix*) {}

    virtual void exitMatrix(DOFMatrix*) {}

    virtual void initVector(DOFVectorBase<double>*) {}

    virtual void exitVector(DOFVectorBase<double>*) {}

    /// Destructor.
    virtual ~BoundaryCondition() {}

    /** \brief
     * Adds the local boundary condition for elInfo to object.
     * The dofIndices and localBound as well as nBasFcts are determined by
     * the calling BoundaryManager.
     */
    virtual void fillBoundaryCondition(DOFMatrix             *matrix,
				       ElInfo                *elInfo,
				       const DegreeOfFreedom *dofIndices,
				       const BoundaryType    *localBound,
				       int                    nBasFcts) {}
  
    /** \brief
     * Adds the local boundary condition for elInfo to vector.
     * The dofIndices and localBound as well as nBasFcts are determined by
     * the calling BoundaryManager.
     */
    virtual void fillBoundaryCondition(DOFVectorBase<double>     *vector, 
				       ElInfo                *elInfo,
				       const DegreeOfFreedom *dofIndices,
				       const BoundaryType    *localBound,
				       int                    nBasFcts) {}
  
    /// Returns the boundary residual for the given element. Called by estimator.
    virtual double boundResidual(ElInfo *elInfo, 
				 DOFMatrix *matrix,
				 const DOFVectorBase<double> *dv) 
    { 
      return 0.0; 
    }

    /** \brief
     * Returns whether the condition must be treated as Dirichlet condition
     * while assemblage.
     */
    virtual bool isDirichlet() 
    { 
      return false; 
    }

    /** \brief
     * In some situations it may be required to set Dirichlet boundary conditions,
     * but not to apply them to the matrix. This is for example the case, if the
     * boundary condition is set to a couple matrix. Then, the boundary conditions
     * must be applied to the couple matrix, but they are set to all matrices in this
     * row (to ensure that there are no other element entries in the Dirichlet boundary
     * condition rows).
     */
    virtual bool applyBoundaryCondition()
    {
      return true;
    }

  protected:
    /** \brief
     * Speciefies for which parts of the boundary the condition holds.
     * This id corresponds to the boundary numbers spcified in the
     * macro file. 
     */
    BoundaryType boundaryType;

    /// FiniteElemSpace for this BoundaryCondition.
    const FiniteElemSpace *rowFESpace;

    /// FiniteElemSpace for this BoundaryCondition.
    const FiniteElemSpace *colFESpace;
  };

}

#endif
