#include "MpCCIAdapter.h"
#include "FiniteElemSpace.h"
#include "SMIAdapter.h"
#include "DOFVector.h"
#include "mpi.h"

namespace AMDiS
{

  MpCCIAdapter::MpCCIAdapter(int                 localComm,
			     int                 meshID,
			     int                 partitionID,
			     bool                useMPI,
			     FiniteElemSpace    *feSpace,
			     int                 elementRegion,
			     int                 surfaceRegion,
			     int                 numQuantities,
			     int                *quantityIDs,
			     int                *quantityDims,
			     DOFVector<double> **dofVectors,
			     int                 numSyncPoints,
			     int                *syncPointIDs,
			     const char         *idString,
			     bool (*elementFct)(ElInfo *elInfo),
			     bool (*surfaceFct)(ElInfo *elInfo, int side))
    : localComm_(localComm),
      feSpace_(feSpace),
      applicationID_(1),
      meshID_(meshID),
      partitionID_(partitionID),
      useMPI_(useMPI)
  {
    FUNCNAME("MpCCIAdapter::MpCCIAdapter()");
    int i;

    mpiRank_ = useMPI_ ? MPI::COMM_WORLD.Get_rank() : 0;
    mpiSize_ = useMPI_ ? MPI::COMM_WORLD.Get_size() : 1;

    TEST_EXIT(surfaceRegion == -1)("not yet for surface regions\n");

    // create SMI adapter 
    smiAdapter_ = new SMIAdapter(applicationID_,
				 meshID_,
				 feSpace_,
				 elementRegion,
				 surfaceRegion,
				 elementFct,
				 surfaceFct);

    // MpCCI initialize
    //if(useMPI_) MPI::COMM_WORLD.Barrier();
    //if(mpiRank_ == 0) {

    int error;

    if (idString) {
      error = CCI_Init_with_id_string(NULL, NULL, idString);
      TEST_EXIT(error == 0)
	("CCI_Init_with_id_string() failed\n");    
    } else {
      error = CCI_Init(NULL, NULL);
      TEST_EXIT(error == 0)("CCI_Init() failed\n");
    }

    //}
    //if(useMPI_) MPI::COMM_WORLD.Barrier();

    //}
    //if(useMPI_) MPI::COMM_WORLD.Barrier();

    localCodeID_ = CCI_MYCODEID;
    int remoteCodeID = (localCodeID_ == 1) ? 2 : 1;

    // define communicator
    int sendQuantities[CCI_MAX_NQUANTITIES];
    int sendMeshes[CCI_MAX_NQUANTITIES];
    int sendNQuantities;
    int recvQuantities[CCI_MAX_NQUANTITIES];
    int recvMeshes[CCI_MAX_NQUANTITIES];
    int recvNQuantities;

    if(localComm == -1) {
      localComm_ = CCI_COMM_RCODE[remoteCodeID];
    } else {
      error = CCI_Comm_info(localComm_,
			     &remoteCodeID_,
			     &remoteComm_,
			     CCI_MAX_NQUANTITIES,
			     sendQuantities,
			     sendMeshes,
			     &sendNQuantities,
			     CCI_MAX_NQUANTITIES,
			     recvQuantities,
			     recvMeshes,
			     &recvNQuantities);

      TEST_EXIT(error)("CCI_Comm_info() failed\n");

      MSG("remote comm id: %d\n", remoteComm_);

      error = CCI_Def_comm(localComm_,
			    remoteCodeID_,
			    remoteComm_,
			    numQuantities,
			    quantityIDs,
			    1,
			    &meshID_);

      TEST_EXIT(error)("CCI_Def_comm() failed\n");
    }

    error = CCI_Def_partition(meshID_, partitionID_);

    TEST_EXIT(error == 0)("CCI_Def_partition() failed\n");
  

    transferMeshToMpCCI();

    for (i = 0; i < numSyncPoints; i++) {
      error = CCI_Sync_point_info(syncPointIDs[i],
				   CCI_MAX_NQUANTITIES,
				   sendQuantities,
				   sendMeshes,
				   &sendNQuantities,
				   CCI_MAX_NQUANTITIES,
				   recvQuantities,
				   recvMeshes,
				   &recvNQuantities);

      TEST_EXIT(error == 0)("CCI_Sync_point_info() failed\n");

      MSG("sync point id: %d\n", syncPointIDs[i]);
      int j;
      for(j = 0; j < sendNQuantities; j++) {
	MSG("send quantity: %d\n", sendQuantities[j]);
      }
      for(j = 0; j < recvNQuantities; j++) {
	MSG("recv quantity: %d\n", recvQuantities[j]);
      }

      error = CCI_Def_sync_point(syncPointIDs[i],
				  sendNQuantities,
				  sendQuantities,
				  sendNQuantities,
				  sendMeshes,
				  recvNQuantities,
				  recvQuantities,
				  recvNQuantities,
				  recvMeshes);

      TEST_EXIT(error == 0)("CCI_Def_sync_point() failed\n");
    }

    // define quantities
    DOFVector<double> **ptr = dofVectors;
    for(i = 0; i < numQuantities; i++) {
      int dim = quantityDims[i];
      smiAdapter_->addQuantity(quantityIDs[i], dim, ptr);
      ptr += dim;
    }

    // close MpCCI setup
    int error = CCI_Close_setup(localCodeID_);

    TEST_EXIT(error == 0)("CCI_Close_setup() failed\n");
  }

  MpCCIAdapter::~MpCCIAdapter()
  {
    delete smiAdapter_;
    int error == CCI_Finalize();
    TEST_EXIT(error == 0)("CCI_Finalize() failed\n");
  }

  void MpCCIAdapter::transferMeshToMpCCI()
  {
    // update SMI data
    smiAdapter_->transferMeshToSMI();

    // read SMI data
    int error;
    int dow = Global::getGeo(WORLD);
    int dim = feSpace_->getMesh()->getDim();
    int numNodes;
    int *nodeIDs;
    double *coords;
    int numElems;
    int *elemIDs;
    int numElemNodeIDs;
    int *elemNodeIDs;
    int elemType;
    int *nodesPerElem;

    error = SMI_Begin_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_Begin_read_transaction() failed\n");

    error = SMI_Get_all_nodes(applicationID_,
			      meshID_,
			      &numNodes,
			      &nodeIDs);
    TEST_EXIT(error == SMI_OK)("SMI_Get_all_nodes() failed\n");

    coords = new double[numNodes * dow];

    error = SMI_Get_nodes(applicationID_,
			  meshID_,
			  numNodes,
			  dow,
			  nodeIDs,
			  coords);
    TEST_EXIT(error == SMI_OK)("SMI_Get_nodes() failed (%d)\n", error);

    error = SMI_Get_all_elems(applicationID_,
			      meshID_,
			      &numElems,
			      &elemIDs);
    TEST_EXIT(error == SMI_OK)("SMI_Get_all_elems() failed\n");

    error = SMI_Get_elems(applicationID_,
			  meshID_,
			  numElems,
			  elemIDs,
			  &numElemNodeIDs,
			  &elemNodeIDs,
			  NULL,
			  NULL);
    TEST_EXIT(error == SMI_OK)("SMI_Get_elems() failed\n");

    switch(dim) {
    case 1:
      elemType = CCI_ELEM_LINE;
      break;
    case 2:
      elemType = CCI_ELEM_TRIANGLE;
      break;
    case 3:
      elemType = CCI_ELEM_TETRAHEDRON;
      break;
    default: ERROR_EXIT("invalid dim\n");
    }

    nodesPerElem = new int[numElems];
    for (int i = 0; i < numElems; i++)
      nodesPerElem[i] = dim + 1;

    error = SMI_End_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_End_read_transaction() failed\n");

 
    // transfer data to MpCCI
    error = CCI_Def_nodes(meshID_,
			  partitionID_,
			  dow,
			  numNodes,
			  numNodes,
			  nodeIDs,
			  CCI_DOUBLE,
			  coords);
    TEST_EXIT(error == 0)("CCI_Def_nodes() failed\n");

    error = CCI_Def_elems(meshID_,
			  partitionID_,
			  numElems,
			  numElems,
			  elemIDs,
			  1,
			  &elemType,
			  nodesPerElem,
			  elemNodeIDs);
    TEST_EXIT(error == 0)("CCI_Def_elems() failed\n");

    // free memory
    delete [] coords;
    delete [] nodesPerElem;
  }

  void MpCCIAdapter::remesh()
  {
    int error;
    error = CCI_Remesh();
    TEST_EXIT(error == 0)("CCI_Remesh() failed\n");

    transferMeshToMpCCI();

    int mode = CCI_TOTAL_REMESH;
    error = CCI_Close_remesh(1, &meshID_, 1, &mode);
    TEST_EXIT(error == 0)("CCI_Close_remesh() failed\n");
  }

  void MpCCIAdapter::checkConvergence(int  myConvergence,
				      int *globalConvergence)
  {
    int error = CCI_Check_convergence(myConvergence, 
				      globalConvergence,
				      CCI_ANY_CODE);
    //localComm_);
    TEST_EXIT(error == 0)("CCI_Check_convergence() failed\n");
  }

  void MpCCIAdapter::putNodes(int quantityID)
  {
    int error;
    int numNodes;
    int dim;
    int *nodeIDs;
    double *values;

    smiAdapter_->transferQuantitiesToSMI(quantityID);

    error = SMI_Begin_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_Begin_read_transaction() failed\n");

    error = SMI_Get_all_nodes(applicationID_,
			      meshID_,
			      &numNodes,
			      &nodeIDs);
    TEST_EXIT(error == SMI_OK)("SMI_Get_all_nodes() failed\n");

    error = SMI_Get_quantity_info(applicationID_,
				  meshID_,
				  quantityID,
				  NULL,
				  NULL,
				  &dim);

    TEST_EXIT(error == SMI_OK)("SMI_Get_quantity_info() failed\n");

    values = new double[dim * numNodes];

    error = SMI_Get_quantity_values(applicationID_,
				    meshID_,
				    quantityID,
				    SMI_TYPE_DOUBLE,
				    dim,
				    numNodes,
				    nodeIDs,
				    values);

    TEST_EXIT(error == SMI_OK)("SMI_Get_quantity_values() failed\n");

    error = SMI_End_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_End_read_transaction() failed\n");

    // copy double to float
    float *floatValues = new float[dim * numNodes];

#if 0
    for (int i = 0; i < dim*numNodes; i++)
      floatValues[i] = static_cast<float>(values[i]);

    error = CCI_Put_nodes(meshID_,
			  partitionID_,
			  quantityID,
			  dim,
			  numNodes,
			  numNodes,
			  nodeIDs,
			  CCI_FLOAT,
			  floatValues);
    TEST_EXIT(error == 0)("CCI_Put_nodes() failed\n");

    delete [] floatValues;
#endif

    error = CCI_Put_nodes(meshID_,
			  partitionID_,
			  quantityID,
			  dim,
			  numNodes,
			  numNodes,
			  nodeIDs,
			  CCI_DOUBLE,
			  values);
    TEST_EXIT(error == 0)("CCI_Put_nodes() failed\n");

    delete [] values;
  }

  void MpCCIAdapter::getNodes(int quantityID)
  {
    int error;
    int numNodes;
    int dim;
    int *nodeIDs;
    double *values;
    int numEmptyNodes;

    error = SMI_Begin_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_Begin_read_transaction() failed\n");

    error = SMI_Get_all_nodes(applicationID_,
			      meshID_,
			      &numNodes,
			      &nodeIDs);
    TEST_EXIT(error == SMI_OK)("SMI_Get_all_nodes() failed\n");

    error = SMI_Get_quantity_info(applicationID_,
				  meshID_,
				  quantityID,
				  NULL,
				  NULL,
				  &dim);

    TEST_EXIT(error == SMI_OK)("SMI_Get_quantity_info() failed (%d)\n", error);

    error = SMI_End_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_End_read_transaction() failed\n");

    values = new double[dim * numNodes];

#if 0
    // copy double to float
    float *floatValues = new float[dim * numNodes];

    error = CCI_Get_nodes(meshID_,
			  partitionID_,
			  quantityID,
			  dim,
			  numNodes,
			  numNodes,
			  nodeIDs,
			  CCI_FLOAT,
			  floatValues,
			  0,
			  NULL,
			  &numEmptyNodes);
    TEST_EXIT(error == 0)("CCI_Get_nodes() failed\n");

    for (int i = 0; i < dim * numNodes; i++) 
      values[i] = static_cast<double>(floatValues[i]);

    delete [] floatValues;
#endif

    error = CCI_Get_nodes(meshID_,
			  partitionID_,
			  quantityID,
			  dim,
			  numNodes,
			  numNodes,
			  nodeIDs,
			  CCI_DOUBLE,
			  values,
			  0,
			  NULL,
			  &numEmptyNodes);
    TEST_EXIT(error == 0)("CCI_Get_nodes() failed\n");


    error = SMI_Begin_write_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_Begin_write_transaction() failed\n");

    error = SMI_Set_quantity_values(applicationID_,
				    meshID_,
				    quantityID,
				    SMI_TYPE_DOUBLE,
				    dim,
				    numNodes,
				    nodeIDs,
				    values);

    TEST_EXIT(error == SMI_OK)("SMI_Set_quantity_values() failed\n");

    error = SMI_End_write_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_End_write_transaction() failed\n");

    smiAdapter_->getQuantitiesFromSMI(quantityID);

    delete [] values;
  }

  void MpCCIAdapter::send(int numQuantities, int *quantityIDs)
  {
    int error = CCI_Send(numQuantities,
			 quantityIDs,
			 1,
			 &meshID_,
			 localComm_);
    TEST_EXIT(error == 0)("CCI_Send() failed\n");
  }

  void MpCCIAdapter::isend(int numQuantities, int *quantityIDs, CCI_Request *request)
  {
    int error = CCI_Isend(numQuantities,
			  quantityIDs,
			  1,
			  &meshID_,
			  localComm_,
			  request);
    TEST_EXIT(error == 0)("CCI_Isend() failed\n");
  }

  void MpCCIAdapter::wait(CCI_Request *request)
  {
    CCI_Status status;
    int error = CCI_Wait(request,
			 &status);
    TEST_EXIT(error == 0)("CCI_Wait() failed\n");  
  }

  void MpCCIAdapter::recv(int numQuantities, int *quantityIDs)
  {
    CCI_Status status;
    MSG("numQuantities %d\n", numQuantities);
    MSG("meshID %d\n", meshID_);
    MSG("localComm %d\n", localComm_);
    int i;
    for(i = 0; i < numQuantities; i++) {
      MSG("quantity %d: %d\n", i, quantityIDs[i]);
    }
    int error = CCI_Recv(numQuantities,
			 quantityIDs,
			 1,
			 &meshID_,
			 localComm_,
			 &status);
    TEST_EXIT(error == 0)("CCI_Send() failed\n");
  }

  void MpCCIAdapter::reachSyncPoint(int syncPointID)
  {
    CCI_Status status;
    int error = CCI_Reach_sync_point(syncPointID, &status);
    TEST_EXIT(error == 0)("CCI_Reach_sync_point() failed\n");
  }

  void MpCCIAdapter::closeCouplingStep()
  {
    int error = CCI_Close_coupling_step(1, 0, 0, 0, 0, 0, 0);
    TEST_EXIT(error == 0)("CCI_Close_coupling_step() failed\n");
  }

  void MpCCIAdapter::setIndexMappings(std::map<DegreeOfFreedom, DegreeOfFreedom> *newNodeIndex,
				      std::map<DegreeOfFreedom, DegreeOfFreedom> *oldNodeIndex,
				      std::map<int, int> *newElementIndex,
				      std::map<int, int> *oldElementIndex)
  {
    smiAdapter_->setIndexMappings(newNodeIndex, 
				  oldNodeIndex,
				  newElementIndex,
				  oldElementIndex);
  }

  int MpCCIAdapter::getNumElements()
  {
    int numElems = 0;
    int *elemIDs;
    int error = SMI_Begin_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_Begin_read_transaction() failed\n");

    error = SMI_Get_all_elems(applicationID_,
			      meshID_,
			      &numElems,
			      &elemIDs);

    TEST_EXIT(error == SMI_OK)("SMI_Get_all_elems() failed\n");

    error = SMI_End_read_transaction(applicationID_, meshID_);
    TEST_EXIT(error == SMI_OK)("SMI_End_read_transaction() failed\n");

    return numElems;
  }

}
