// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file QPInfo.h */

#ifndef AMDIS_QPINFO_H
#define AMDIS_QPINFO_H

#include <map>
#include "AMDiS_fwd.h"
#include "FixVec.h"

namespace AMDiS {

  /** \brief
   * Stores informations at quadrature points of the current element.
   */
  class QPInfo 
  {
  public:
    /** \brief
     * Sets \ref currentElInfo_ to elInfo and all valid flags to false.
     */
    void initElement(const ElInfo *elInfo);

    /** \brief
     * Returns coordinates at quadrature points.
     */
    WorldVector<double> *getCoordsAtQPs(int numPoints);

    /** \brief
     * Returns vector values at quadrature points. If quadFast is set it will be 
     * used for a more efficient evaluation.
     */
    double *getVecAtQPs(const DOFVector<double>*, 
			int numPoints,
			const FastQuadrature *quadFast = NULL);

    /** \brief
     * Returns gradient values at quadrature points. If quadFast is set it will be 
     * used for a more efficient evaluation.
     */
    WorldVector<double> *getGrdAtQPs(const DOFVector<double>*, 
				     int numPoints,
				     const FastQuadrature *quadFast = NULL);

    /** \brief
     * Returns D2 values at quadrature points. If quadFast is set it will be 
     * used for a more efficient evaluation.
     */
    WorldMatrix<double> *getD2AtQPs(const DOFVector<double>*, 
				    int numPoints,
				    const FastQuadrature *quadFast = NULL);


    /** \brief
     * Returns element normals at quadrature points.
     */
    WorldVector<double> **getElementNormalAtQPs(int numPoints);

    /** \brief
     * 
     */
    DimVec<WorldVector<double> > **getGrdLambdaAtQPs(int numPoints);

    /** \brief
     * Returns a QPInfo instance for the given quadrature.
     */
    static QPInfo *provideQPInfo(const Quadrature*, const FastQuadrature*);

    /** \brief
     * Deletes the QPInfo instance for the given quadrature.
     */
    static void clearQPInfo(const Quadrature*, const FastQuadrature*);

    /** \brief
     * Deletes all QPInfo instances.
     */
    static void clearAllQPInfos();

  protected:
    /** \brief
     * Constructor. Called by \ref provideQPInfo().
     */
    QPInfo(const Quadrature*);

    /** \brief
     * Destructor. Called by \ref clearQPInfo() and \ref clearAllQPInfos().
     */
    ~QPInfo();

  protected:
    /** \brief
     * Structure which stores infos about one DOFVector.
     */
    class VecQPInfo 
    {
    public:
      /** \brief
       * Constructor.
       */
      VecQPInfo() 
	: valAtQPs_(NULL),
	  grdAtQPs_(NULL),
	  D2AtQPs_(NULL),
	  valNumPointsValid_(0),
	  grdNumPointsValid_(0),
	  D2NumPointsValid_(0)
      {}

      /** \brief
       * Values at quadrature points
       */
      double *valAtQPs_;

      /** \brief
       * Gradients at quadrature points
       */
      WorldVector<double> *grdAtQPs_;
    
      /** \brief
       * D2 at quadrature points
       */
      WorldMatrix<double> *D2AtQPs_;

      /** \brief
       * valid flag for values
       */
      int valNumPointsValid_;

      /** \brief
       * valid flag for gradients
       */
      int grdNumPointsValid_;

      /** \brief
       * valid flag for D2
       */
      bool D2NumPointsValid_;
    };

    /** \brief
     * Quadrature of this QPInfo
     */
    const Quadrature *quadrature_;

    /** \brief
     * Set to \ref quadrature_->getNumPoints().
     */
    int numPoints_;

    /** \brief
     * ElInfo of the current element
     */
    const ElInfo *currentElInfo_;

    /** \brief
     * Coords at quadrature points
     */
    WorldVector<double> *coordsAtQPs_;

    /** \brief
     * Valid flag for coords
     */
    int coordsNumPointsValid_;

    /** \brief
     * Map of all vector infos
     */
    std::map<const DOFVector<double>*, VecQPInfo*> vecQPInfos_;

    /** \brief
     * element normal at quadrature points (array of pointers)
     */
    WorldVector<double> **elementNormalAtQPs_;

    /** \brief
     * for constant values at all QPs (all entries point to same memory)
     */
    WorldVector<double> **elementNormalConst_;
  
    /** \brief
     * valid flag for element normals
     */
    int elementNormalNumPointsValid_;

    /** \brief
     * gradient of barycentric coordinates at QPs (array of pointers)
     */
    DimVec<WorldVector<double> > **grdLambdaAtQPs_;

    /** \brief
     * for constant values at all QPs (all entries point to same memory)
     */
    DimVec<WorldVector<double> > **grdLambdaConst_;

    /** \brief
     * number of valid points of grdLambdaAtQPs_
     */
    int grdLambdaNumPointsValid_;

    /** \brief
     * Static map of all QPInfos. Used by \ref provideQPInfo().
     */
    static std::map<const Quadrature*, QPInfo*> qpInfos_;
  };

}

#endif
