// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ResidualEstimator.h */

/** \defgroup Estimator Estimator module
 * @{ <img src="estimator.png"> @}
 */

#ifndef AMDIS_RESIDUALESTIMATOR_H
#define AMDIS_RESIDUALESTIMATOR_H

#include "Estimator.h"
#include "FixVec.h"

namespace AMDiS {

  /** \brief
   * Returns residual square at quadrature point. Not Member of
   * Estimator to avoid multiple instantiation.
   */
  void r(const ElInfo *elInfo,
	 int nPoints,
	 const double *uhIq,
	 const WorldVector<double> *grdUhIq,
	 const WorldMatrix<double> *D2UhIq,
	 const double *uhOldIq,
	 const WorldVector<double> *grdUhOldIq,
	 const WorldMatrix<double> *D2UhOldIq,
	 DOFMatrix *A, 
	 DOFVector<double> *fh,
	 Quadrature *quad,
	 double *result);
 
  /// Returns pow(det,2.0/dim). Not Member of Estimator to avoid multiple instantiation.
  inline double h2_from_det(double det, int dim) 
  {
    return pow(det, 2.0 / dim);
  }

  /**
   * \ingroup Estimator
   * 
   * \brief
   * Estimator for scalar problems.
   */
  class ResidualEstimator : public Estimator
  {
  public:
    /// Creator class used in the OEMSolverMap.
    class Creator : public EstimatorCreator
    {
    public:
      Creator() : EstimatorCreator() {}

      virtual ~Creator() {}

      /// Returns a new ODirSolver object.
      Estimator* create() 
      { 
	return new ResidualEstimator(name, row);
      }
    };
  
    /// Constructor.
    ResidualEstimator(std::string name, int r);

    virtual void init(double timestep);

    virtual void estimateElement(ElInfo *elInfo);

    virtual void exit(bool output = true);

  protected:
    /// Constant in front of element residual
    double C0;
 
    /// Constant in front of edge/face residual
    double C1;

    /// Not used! Was thought to be the constant in front of coarsening term.
    double C2;

    /// Constant in front of the time
    double C3;

    /// Number of systems, e.g., number of variables in the equation.
    int nSystems;

    /// Number of quadrature points.
    int nPoints;

    int dim;

    int degree;

    Quadrature *quad;

    FastQuadrature **quadFast;

    const BasisFunction **basFcts;

    double **uhEl;

    double **uhOldEl;

    double **uhNeigh;

    double *uhQP;

    double *uhOldQP;

    /// Stores the element residual computed at the quadrature points of the element.
    double *riq;

    WorldVector<double> *grdUh_qp;

    WorldMatrix<double> *D2uhqp; 

    ElInfo *neighInfo;

    Quadrature *surfaceQuad;

    int nPointsSurface;

    Vector<WorldVector<double> > grdUhEl;

    Vector<WorldVector<double> > grdUhNeigh;

    Vector<WorldVector<double> > jump;

    Vector<WorldVector<double> > localJump;

    WorldVector<int> faceIndEl;
    
    WorldVector<int> faceIndNeigh;

    DimVec<WorldVector<double> > *lambdaNeigh;

    DimVec<double> *lambda;

    /// Maximal number of neighbours an element may have in the used dimension.
    int nNeighbours;
  };
}

#endif // AMDIS_RESIDUALESTIMATOR_H
