#include "ResidualParallelEstimator.h"
#include "ResidualEstimator.h"
#include "TraverseParallel.h"
#include "Parameters.h"
#include "OpenMP.h"

#ifdef _OPENMP

namespace AMDiS {
  ResidualParallelEstimator::ResidualParallelEstimator(std::string name, int r)
    : Estimator(name, r),
      C3(1.0)
  {
    GET_PARAMETER(0, name + "->C3", "%f", &C3);
    C3 = C3 > 1.e-25 ? sqr(C3) : 0.0;

    seqEstimators_.resize(omp_get_overall_max_threads());

    for (int i = 0; i < omp_get_overall_max_threads(); i++)
      seqEstimators_[i] = new ResidualEstimator(name, r);
  }

  ResidualParallelEstimator::~ResidualParallelEstimator()
  {
    for (int i = 0; i < static_cast<int>(seqEstimators_.size()); i++)
      delete seqEstimators_[i];    
  }

  void ResidualParallelEstimator::addSystem(DOFMatrix *matrix_,
					    DOFVector<double> *uh_,
					    DOFVector<double> *fh_,
					    DOFVector<double> *uhOld_)
  {
    Estimator::addSystem(matrix_, uh_, fh_, uhOld_);

    for (int i = 0; i < static_cast<int>(seqEstimators_.size()); i++)
      seqEstimators_[i]->addSystem(matrix_, uh_, fh_, uhOld_);
  }

  void ResidualParallelEstimator::addUhOldToSystem(int system, 
						   DOFVector<double> *uhOld_)
  {
    Estimator::addUhOldToSystem(system, uhOld_);

    for (int i = 0; i < static_cast<int>(seqEstimators_.size()); i++)
      seqEstimators_[i]->addUhOldToSystem(system, uhOld_);
  }

  double ResidualParallelEstimator::estimate(double ts)
  {
    FUNCNAME("ResidualParallelEstimator::estimate()");

    mesh = uh[row == -1 ? 0 : row]->getFESpace()->getMesh();

    for (int i = 0; i < static_cast<int>(seqEstimators_.size()); i++)
      seqEstimators_[i]->init(ts);

    TraverseParallelStack stack;

#pragma omp parallel
    {
      ResidualEstimator *estimator = seqEstimators_[omp_get_thread_num()];     
   
      traverseFlag = 
	Mesh::FILL_NEIGH      |
	Mesh::FILL_COORDS     |
	Mesh::FILL_OPP_COORDS |
	Mesh::FILL_BOUND      |
	Mesh::FILL_GRD_LAMBDA |
	Mesh::FILL_DET        |
	Mesh::CALL_LEAF_EL;
   
      ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);

      while (elInfo) {
	estimator->estimateElement(elInfo);
	elInfo = stack.traverseNext(elInfo);
      }
    }

    est_sum = 0.0;
    est_max = 0.0;
    est_t_sum = 0.0;
    est_t_max = 0.0;

    for (int i = 0; i < static_cast<int>(seqEstimators_.size()); i++) {
      seqEstimators_[i]->exit(false);
      est_sum += pow(seqEstimators_[i]->getErrorSum(), 2);
      est_max = max(est_max, seqEstimators_[i]->getErrorMax());
      est_t_sum += pow(seqEstimators_[i]->getTimeEst(), 2);
      est_t_max = max(est_t_max, seqEstimators_[i]->getTimeEstMax());
    }

    est_sum = sqrt(est_sum);
    est_t_sum = sqrt(est_t_sum);

    MSG("estimate   = %.8e\n", est_sum);
    if (C3)
      MSG("time estimate   = %.8e\n", est_t_sum);

    return est_sum;
  }


}

#endif // _OPENMP
