// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file SurfaceOperator.h */

#ifndef AMDIS_SURFACEOPERATOR_H
#define AMDIS_SURFACEOPERATOR_H

#include "FiniteElemSpace.h"
#include "Mesh.h"
#include "Operator.h"
#include "SurfaceQuadrature.h"
#include "OpenMP.h"

namespace AMDiS {

  /** 
   * \ingroup Integration
   *
   * \brief
   * A SurfaceOperator is a Operator used for surface integration. Instead of a 
   * normal Quadrature it uses a SurfaceQuadrature with integration points at
   * one element side instead of integration points in the inner of the element.
   * The determinant in ElInfo is replaced by the surface determinant, so the
   * assemblage can be done with the standard Assembler classes.
   * The SurfaceQuadrature is used for the implementation of Neumann and Robin
   * boundary conditions.
   */
  class SurfaceOperator : public Operator
  {
  public:
    /// Creates a SurfaceOperator conforming to operat for the given \ref coords.
    SurfaceOperator(Operator *operat, 
		    VectorOfFixVecs<DimVec<double> > &coords) 
      : Operator(*operat), 
	coords_(coords),
	quad2(NULL),
	quad1GrdPsi(NULL),
	quad1GrdPhi(NULL),
	quad0(NULL)
    {
      assembler[omp_get_thread_num()] = NULL;

      int dim = rowFESpace->getMesh()->getDim();
      int degree;
      int myRank = omp_get_thread_num();

      // create surface quadratures
      if (secondOrder[myRank].size() > 0) {
	degree = getQuadratureDegree(2);
	// quad2 = SurfaceQuadrature::provideSurfaceQuadrature(dim, degree, coords);
	quad2 = new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree),
				      coords);
      }

      if (firstOrderGrdPsi[myRank].size() > 0) {
	degree = getQuadratureDegree(1, GRD_PSI);
	quad1GrdPsi = new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree),
					    coords);
	//SurfaceQuadrature::provideSurfaceQuadrature(dim, degree, coords);
      }

      if (firstOrderGrdPhi[myRank].size() > 0) {
	degree = getQuadratureDegree(1, GRD_PHI);
	quad1GrdPhi = new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree),
					    coords);
	//SurfaceQuadrature::provideSurfaceQuadrature(dim, degree, coords);
      }

      if (zeroOrder[myRank].size() > 0) {
	degree = getQuadratureDegree(0);
	quad0 = new SurfaceQuadrature(Quadrature::provideQuadrature(dim - 1, degree),
				      coords);
	//SurfaceQuadrature::provideSurfaceQuadrature(dim, degree, coords);
      }

      // initialize assembler with surface quadratures
      optimized = false;
      initAssembler(myRank, quad2, quad1GrdPsi, quad1GrdPhi, quad0);
    }

    /// Adapt surface quadratures to \ref coords.
    void adaptSurfaceOperator(VectorOfFixVecs<DimVec<double> > &coords)
    {
      coords_ = coords;

      if (quad2)
	quad2->scaleSurfaceQuadrature(coords);
      if (quad1GrdPsi)
	quad1GrdPsi->scaleSurfaceQuadrature(coords);
      if (quad1GrdPhi)
	quad1GrdPhi->scaleSurfaceQuadrature(coords);
      if (quad0)
	quad0->scaleSurfaceQuadrature(coords);
    }

    /** \brief
     * Implementation of \ref Operator::getElementMatrix(). Repalces the 
     * determinant by the surface determinant and deligates the call to
     * the base class function.
     */
    virtual void getElementMatrix(const ElInfo *elInfo, 
				  ElementMatrix& userMat, 
				  double factor = 1.0)
    {
      int dim = rowFESpace->getMesh()->getDim();
      double origDet = elInfo->getDet();

      FixVec<WorldVector<double>, VERTEX> worldCoords(dim-1, NO_INIT);

      // transform barycentric coords to world coords
      for (int i = 0; i < dim; i++)
	elInfo->coordToWorld(coords_[i], worldCoords[i]);

      // set determinant for world coords of the side
      const_cast<ElInfo*>(elInfo)->setDet(elInfo->calcDet(worldCoords));

      // calc element matrix
      Operator::getElementMatrix(elInfo, userMat, factor);

      // set determinant for world coords of the side
      const_cast<ElInfo*>(elInfo)->setDet(origDet);
    }
  
    /** \brief
     * Implementation of \ref Operator::getElementVector(). Repalces the 
     * determinant by the surface determinant and deligates the call to
     * the base class function.
     */
    virtual void getElementVector(const ElInfo *elInfo, 
				  ElementVector& userVec, 
				  double factor = 1.0) 
    {
      int dim = rowFESpace->getMesh()->getDim();
      double origDet = elInfo->getDet();

      FixVec<WorldVector<double>, VERTEX> worldCoords(dim-1, NO_INIT);

      // transform barycentric coords to world coords
      for (int i = 0; i < dim; i++)
	elInfo->coordToWorld(coords_[i], worldCoords[i]);

      // set determinant for world coords of the side
      const_cast<ElInfo*>(elInfo)->setDet(elInfo->calcDet(worldCoords));

      // calc element vector
      Operator::getElementVector(elInfo, userVec, factor);

      const_cast<ElInfo*>(elInfo)->setDet(origDet);
    }

  protected:
    VectorOfFixVecs<DimVec<double> > coords_;

    /// Surface quadratures
    SurfaceQuadrature *quad2;
    SurfaceQuadrature *quad1GrdPsi;
    SurfaceQuadrature *quad1GrdPhi;
    SurfaceQuadrature *quad0;
  };

}

#endif
